#!/usr/bin/env python

__doc__ = """
    Executes Metaphor on real data.
    """


import sys
from argparse import Namespace
from pathlib import Path
from shutil import copyfile
from subprocess import CalledProcessError, check_call

from metaphor import wrapper_prefix
from metaphor.workflow import snakefile
from metaphor.config import default_config
from metaphor.utils import confirm_message, get_successful_completion, load_yaml

from .create_input_table import main as create_input_table


def main(args):

    # Initial sanity check
    config_file = args.configfile
    input_dir = args.input_dir
    join_units = args.join_units
    cores = args.cores
    max_mb = args.max_mb
    confirm = args.confirm
    extras = args.extras
    profile = args.profile

    if not Path(config_file).exists():
        if not confirm:
            yn = input(
                f"Config file '{config_file}' does not exist yet. Metaphor will create one based on the default config contained on: '{default_config}'.\n"
                "This does not include all of Metaphor's features. We recommend you create your own config file with the 'metaphor config settings' command.\n"
                "Ok to continue? [y/N]\n"
            )
        if yn.lower() == "y":
            print(f"Copying default config {config_file} to current directory.")
            copyfile(default_config, config_file)
        else:
            print("Metaphor execution cancelled.")
            sys.exit()

    # Load config file to check for input_dir
    config = load_yaml(config_file)

    samples = config["samples"]
    if max_mb is None:
        max_mb = config["max_mb"]
    if not input_dir:
        assert Path(
            samples
        ).exists(), f"Samples file {samples} doesn't exist. Please provide an input directory with FASTQ or a valid samples file."
    else:
        create_input_table_args = Namespace()
        create_input_table_args.input_dir = input_dir
        create_input_table_args.output_file = samples = "samples.csv"
        create_input_table_args.join_units = join_units
        print("\nCreating input table for test files.\n")
        create_input_table(create_input_table_args)

    print(f"Your config file is '{config_file}'.\n")

    print("Starting Snakemake.")
    print(
        "This may require the installation of conda environments which should take a while.\n"
    )
    if not confirm:
        confirm_message(cores, max_mb)

    cmd = f"""
    snakemake --snakefile {snakefile}           \
              --configfile {config_file}        \
              --cores {cores}                   \
              -p -r                             \
              --use-conda                       \
              --wrapper-prefix {wrapper_prefix}
    """

    for arg in ("samples", "max_mb"):
        if value := eval(arg):
            cmd += f" --config {arg}={value} "

    if profile:
        cmd += f" --profile {profile} "

    cmd += f" {extras} "

    try:
        retcode = check_call(cmd.split())
    except CalledProcessError as e:
        retcode = e.returncode
        raise
    except Exception as e:
        retcode = 1
        raise
    get_successful_completion(retcode, "Metaphor finished successfully.")
